from typing import List, Optional as Opt

from spoffy.models.image import Image
from spoffy.models.paging import OffsetPaging
from spoffy.models.base import SpotifyObject


class Copyright(SpotifyObject):
    text: str
    type: str


class ExternalUrls(SpotifyObject):
    spotify: str


class Restrictions(SpotifyObject):
    reason: str


class ExternalIds(SpotifyObject):
    upc: Opt[str] = None


class TrackExternalIds(SpotifyObject):
    isrc: str


class Followers(SpotifyObject):
    href: Opt[str]
    total: int


class ArtistSimple(SpotifyObject):
    id: str
    external_urls: ExternalUrls
    href: str
    name: str
    type: str
    uri: str


class Artist(ArtistSimple):
    followers: Followers
    genres: List[str]
    images: List[Image]
    popularity: int


class TrackLink(SpotifyObject):
    external_urls: ExternalUrls
    href: str
    id: str
    type: str
    uri: str


class TrackSimple(SpotifyObject):
    id: str
    name: str
    artists: List[ArtistSimple]
    disc_number: int
    duration_ms: int
    explicit: bool
    external_urls: ExternalUrls
    href: str
    preview_url: Opt[str]
    track_number: int
    type: str
    is_local: bool
    uri: str
    available_markets: Opt[List[str]] = None
    linked_from: Opt[TrackLink] = None
    is_playable: Opt[bool] = None


class AlbumBase(SpotifyObject):
    id: str
    name: str
    album_type: str
    artists: List[ArtistSimple]
    external_urls: ExternalUrls
    href: str
    images: List[Image]
    release_date: str
    release_date_precision: str
    type: str
    uri: str
    total_tracks: int


class AlbumSimple(AlbumBase):
    restrictions: Opt[Restrictions] = None
    available_markets: Opt[List[str]] = None


class Track(TrackSimple):
    explicit: bool
    album: AlbumSimple
    popularity: int
    external_ids: TrackExternalIds
    available_markets: Opt[List[str]] = None
    linked_from: Opt[TrackLink] = None
    is_playable: Opt[bool] = None


class AlbumTracksPaging(OffsetPaging):
    items: List[TrackSimple]


class Album(AlbumBase):
    copyrights: List[Copyright]
    external_ids: ExternalIds
    genres: List[str]
    label: str
    popularity: int
    tracks: AlbumTracksPaging
    restrictions: Opt[Restrictions] = None
    available_markets: Opt[List[str]] = None


class ArtistAlbumsPaging(OffsetPaging):
    items: List[AlbumSimple]



from typing import List, Optional as Opt

from spoffy.models.core import ExternalUrls, Followers
from spoffy.models.image import Image
from spoffy.models.base import SpotifyObject


class UserBase(SpotifyObject):

    external_urls: ExternalUrls
    href: str
    id: str
    type: str
    uri: str


class UserPublic(UserBase):
    followers: Followers
    images: List[Image]
    display_name: Opt[str]


class UserPrivate(UserPublic):
    birthdate: Opt[str] = None
    country: Opt[str] = None
    email: Opt[str] = None
    product: Opt[str] = None



import json
from pprint import pformat
from typing import Dict, Any, List

from cattr import Converter


class SpotifyObject:
    __spotify_object__ = True

    def __init__(self, **d: Dict[str, Any]) -> None:

        # type_hints = getattr(self.__class__, "__annotations__", {})
        self.__annos = getattr(self.__class__, "__annotations__", {})
        self.__attrs: List[str] = []
        for key, value in d.items():
            self.__attrs.append(key)
            if isinstance(value, dict):
                setattr(self, key, self.__convert_nested(key, value))
            elif isinstance(value, list) and value:
                if isinstance(value[0], dict):
                    setattr(self, key, self.__convert_nested_list(key, value))
                else:
                    setattr(self, key, value)
            else:
                setattr(self, key, value)

    def __convert_nested(self, key, value: dict):
        return converter.structure(value, self.__annos.get(key, SpotifyObject))

    def __convert_nested_list(self, key, value: List[dict]):
        return converter.structure(
            value, self.__annos.get(key, List[SpotifyObject])
        )

    def to_dict(self) -> Dict[str, Any]:
        result = {}
        for key in self.__attrs:
            value = getattr(self, key)
            if hasattr(value, "to_dict"):
                value = value.to_dict()
            elif isinstance(value, list) and value:
                dict_list = hasattr(value[0], "to_dict")
                if dict_list:
                    value = [item.to_dict() for item in value]
            result[key] = value
        return result

    def __str__(self):
        return json.dumps(self.to_dict(), indent=2)

    def __repr__(self):
        return "<{} ({})>".format(
            self.__class__.__name__, pformat(self.to_dict())
        )


def convert_spotobject(obj, cl):
    return cl(**obj)


def is_obj(cls):
    return getattr(cls, "__spotify_object__", False)


converter = Converter()
converter.register_structure_hook_func(is_obj, convert_spotobject)
# converter.register_structure_hook(SpotifyObject, convert_spotobject)



import sys
from typing import Sequence

version_info = sys.version_info[0:3]
is_py37 = version_info[:2] == (3, 7)

if is_py37:
    from typing import List, Union, _GenericAlias  # type: ignore

    def is_union(obj):
        return (
            obj is Union
            or isinstance(obj, _GenericAlias)
            and obj.__origin__ is Union
        )

    def is_list(type):
        return (
            type is List
            or type.__class__ is _GenericAlias
            and issubclass(type.__origin__, List)
        )

    def is_sequence(type):
        return (
            type is List
            or type.__class__ is _GenericAlias
            and issubclass(type.__origin__, Sequence)
        )


else:
    from typing import _Union, List  # type: ignore

    def is_union(obj):
        """Return true if the object is a union. """
        return isinstance(obj, _Union)

    def is_list(type):
        return issubclass(type, List)

    def is_sequence(type):
        return issubclass(type, Sequence)



from typing import Optional as Opt, Dict


class SpotifyException(Exception):
    """
    Base exception for any Spotify API errors

    :param message: The error message from the API
    :param status_code: The HTTP status code of the error response
    :param url: The URL of the failed request
    :param request_method: The HTTP method of the failed request
    :param reason: The error reason as returned from the API
        (not included on all errors)
    :param error_description: A more detailed error description as
        returned from the API (not included on all errors)
    """

    def __init__(
        self,
        message: str,
        *,
        request_method: str,
        request_url: str,
        status_code: int,
        headers: Dict[str, str],
        reason: Opt[str] = None,
        error_description: Opt[str] = None,
    ):
        self.message = message
        self.status_code = status_code
        self.headers = headers
        self.reason = reason
        self.error_description = error_description
        self.request_method = request_method
        self.request_url = request_url

    def __str__(self):
        return "{status_code}: {message} ({method} {url})".format(
            status_code=self.status_code,
            message=self.message,
            method=self.request_method,
            url=self.request_url,
        )


class SpotifyUnauthorized(SpotifyException):
    """
    Raised on authorization errors such
    as when an invalid access token is used
    """


class SpotifyPremiumRequired(SpotifyException):
    """
    Raise when attempting to call and endpoint
    which requires a premium subscription with
    a free account access token
    """



from typing import Optional as Opt, List, Any

from spoffy.models.base import SpotifyObject


class Paging(SpotifyObject):
    href: str
    limit: int
    next: Opt[str]
    previous: Opt[str]
    items: List[Any]


class OffsetPaging(Paging):
    """
    Offset based paging object

    :param href: Url to this object
    :param href: URL to this object
    :param limit: Max number of items per page
    :param offset: The paging offset
    :param next: URL to next page (if available)
    :param previous: URL to previous page (if available)
    """

    # Jedi doesn't support generics, do this for now
    href: str
    limit: int
    offset: int
    total: int
    next: Opt[str]
    previous: Opt[str]


class Cursor(SpotifyObject):
    after: str
    before: str


class CursorPaging(Paging):
    """
    Cursor based paging object

    :param href: URL to this object
    :param limit: Max number of items per page
    :param cursors: Cursor object for pagination
    :param next: URL to next page (if available)
    :param previous: URL to previous page (if available)
    """

    href: str
    limit: int
    cursors: Cursor
    next: Opt[str] = None
    previous: Opt[str] = None



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['spoffy', 'spoffy.client', 'spoffy.io', 'spoffy.models', 'spoffy.modules']

package_data = \
{'': ['*']}

install_requires = \
['cattrs>=0.9,<0.10', 'urlobject>=2,<3']

setup_kwargs = {
    'name': 'spoffy',
    'version': '0.3.1',
    'description': 'Spotify API client with async and sync support',
    'long_description': '# Spoffy\n\n[![CircleCI](https://circleci.com/gh/steinitzu/spoffy/tree/master.svg?style=svg)](https://circleci.com/gh/steinitzu/spoffy/tree/master)\n[![Documentation Status](https://readthedocs.org/projects/spoffy/badge/?version=latest&style=flat-square)](https://spoffy.readthedocs.io/en/latest/?badge=latest)\n\n\nThe IDE friendly sync and async [Spotify API](https://developer.spotify.com/documentation/web-api/) wrapper for python.\n\nRead the docs: https://spoffy.readthedocs.io\n\n\n# Install\n\n```\npip install spoffy\n```\nPython3.6 and newer are supported\n\n\n\n',
    'author': 'Steinthor Palsson',
    'author_email': 'steini90@gmail.com',
    'url': 'https://github.com/steinitzu/spoffy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
